@@ -31,6 +31,8 @@
 #include "editor-application-private.h"
 #include "editor-buffer-monitor-private.h"
 #include "editor-document-private.h"
+#include "editor-document-statistics.h"
+#include "editor-encoding-model.h"
 #include "editor-session-private.h"
 #include "editor-window-private.h"
 
@@ -51,9 +53,14 @@ struct _EditorDocument
   const GtkSourceEncoding      *encoding;
   GError                       *last_error;
 
+  EditorDocumentStatistics     *statistics;
+
   SpellingChecker              *spell_checker;
   SpellingTextBufferAdapter    *spell_adapter;
 
+  GtkTextMark                  *saved_insert_mark;
+  GtkTextMark                  *saved_selection_bound_mark;
+
   GtkSourceNewlineType          newline_type;
   guint                         busy_count;
   gdouble                       busy_progress;
@@ -66,6 +73,8 @@ struct _EditorDocument
   guint                         suggest_admin : 1;
   guint                         load_failed : 1;
   guint                         did_shutdown : 1;
+  guint                         inserting : 1;
+  guint                         deleting : 1;
 };
 
 typedef struct
@@ -99,18 +108,23 @@ enum {
   PROP_0,
   PROP_BUSY,
   PROP_BUSY_PROGRESS,
+  PROP_ENCODING,
+  PROP_ERROR,
   PROP_EXTERNALLY_MODIFIED,
   PROP_FILE,
   PROP_HAD_ERROR,
   PROP_ERROR_MESSAGE,
   PROP_LOADING,
+  PROP_NEWLINE_TYPE,
   PROP_SPELL_CHECKER,
+  PROP_STATISTICS,
   PROP_SUGGEST_ADMIN,
   PROP_TITLE,
   N_PROPS
 };
 
 enum {
+  CURSOR_JUMPED,
   SAVE,
   N_SIGNALS
 };
@@ -167,6 +181,7 @@ editor_document_track_error (EditorDocument *self,
   g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_SUGGEST_ADMIN]);
   g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_ERROR_MESSAGE]);
   g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_HAD_ERROR]);
+  g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_ERROR]);
 }
 
 static void
@@ -294,6 +309,9 @@ editor_document_load_notify_completed_cb (EditorDocument *self,
 
   self->loading = FALSE;
 
+  if (self->statistics != NULL)
+    editor_document_statistics_reload (self->statistics);
+
   if (self->spell_adapter != NULL)
     spelling_text_buffer_adapter_invalidate_all (self->spell_adapter);
 
@@ -342,6 +360,9 @@ editor_document_changed (GtkTextBuffer *buffer)
     {
       /* Track separately from :modified for drafts */
       self->needs_autosave = TRUE;
+
+      if (self->statistics != NULL)
+        editor_document_statistics_queue_reload (self->statistics);
     }
 
   GTK_TEXT_BUFFER_CLASS (editor_document_parent_class)->changed (buffer);
@@ -418,14 +439,18 @@ editor_document_insert_text (GtkTextBuffer *buffer,
    */
   if (self->loading)
     {
+      self->inserting = TRUE;
       GTK_TEXT_BUFFER_CLASS (editor_document_parent_class)->insert_text (buffer, pos, new_text, new_text_length);
+      self->inserting = FALSE;
       return;
     }
 
   line = gtk_text_iter_get_line (pos);
   offset = gtk_text_iter_get_offset (pos);
 
+  self->inserting = TRUE;
   GTK_TEXT_BUFFER_CLASS (editor_document_parent_class)->insert_text (buffer, pos, new_text, new_text_length);
+  self->inserting = FALSE;
 
   if (offset < TITLE_MAX_LEN && editor_document_get_file (self) == NULL)
     g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_TITLE]);
@@ -453,13 +478,17 @@ editor_document_delete_range (GtkTextBuffer *buffer,
    */
   if (self->loading)
     {
+      self->deleting = TRUE;
       GTK_TEXT_BUFFER_CLASS (editor_document_parent_class)->delete_range (buffer, start, end);
+      self->deleting = FALSE;
       return;
     }
 
   offset = gtk_text_iter_get_offset (start);
 
+  self->deleting = TRUE;
   GTK_TEXT_BUFFER_CLASS (editor_document_parent_class)->delete_range (buffer, start, end);
+  self->deleting = FALSE;
 
   if (offset < TITLE_MAX_LEN && editor_document_get_file (self) == NULL)
     g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_TITLE]);
@@ -551,16 +580,59 @@ apply_spellcheck_mapping (GValue   *value,
   return TRUE;
 }
 
+static void
+editor_document_cursor_moved (EditorDocument *self)
+{
+  g_assert (EDITOR_IS_DOCUMENT (self));
+
+  if (self->inserting || self->deleting)
+    return;
+
+  g_signal_emit (self, signals[CURSOR_JUMPED], 0);
+}
+
+EditorEncoding *
+editor_document_dup_encoding (EditorDocument *self)
+{
+  EditorEncoding *encoding;
+
+  if ((encoding = editor_encoding_model_get (NULL, self->encoding)))
+    return g_object_ref (encoding);
+
+  return NULL;
+}
+
+void
+editor_document_set_encoding (EditorDocument *self,
+                              EditorEncoding *encoding)
+{
+  const GtkSourceEncoding *enc;
+
+  if (encoding != NULL)
+    enc = editor_encoding_get_encoding (encoding);
+  else
+    enc = gtk_source_encoding_get_utf8 ();
+
+  _editor_document_set_encoding (self, enc);
+}
+
 static void
 editor_document_constructed (GObject *object)
 {
   EditorDocument *self = (EditorDocument *)object;
+  GtkTextIter iter;
 
   if (shared_settings == NULL)
     shared_settings = g_settings_new ("org.gnome.TextEditor");
 
   G_OBJECT_CLASS (editor_document_parent_class)->constructed (object);
 
+  gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (self), &iter);
+  self->saved_insert_mark = gtk_text_buffer_create_mark (GTK_TEXT_BUFFER (self),
+                                                         NULL, &iter, TRUE);
+  self->saved_selection_bound_mark = gtk_text_buffer_create_mark (GTK_TEXT_BUFFER (self),
+                                                                  NULL, &iter, FALSE);
+
   self->spell_adapter = spelling_text_buffer_adapter_new (GTK_SOURCE_BUFFER (self),
                                                           self->spell_checker);
   g_settings_bind_with_mapping (shared_settings, "spellcheck",
@@ -574,6 +646,7 @@ editor_document_finalize (GObject *object)
 {
   EditorDocument *self = (EditorDocument *)object;
 
+  g_clear_object (&self->statistics);
   g_clear_object (&self->monitor);
   g_clear_object (&self->file);
   g_clear_object (&self->spell_checker);
@@ -615,6 +688,14 @@ editor_document_get_property (GObject    *object,
       g_value_set_double (value, editor_document_get_busy_progress (self));
       break;
 
+    case PROP_ENCODING:
+      g_value_take_object (value, editor_document_dup_encoding (self));
+      break;
+
+    case PROP_ERROR:
+      g_value_set_boxed (value, self->last_error);
+      break;
+
     case PROP_EXTERNALLY_MODIFIED:
       g_value_set_boolean (value, editor_document_get_externally_modified (self));
       break;
@@ -627,6 +708,14 @@ editor_document_get_property (GObject    *object,
       g_value_set_boolean (value, _editor_document_get_loading (self));
       break;
 
+    case PROP_NEWLINE_TYPE:
+      g_value_set_enum (value, self->newline_type);
+      break;
+
+    case PROP_STATISTICS:
+      g_value_take_object (value, editor_document_load_statistics (self));
+      break;
+
     case PROP_TITLE:
       g_value_take_string (value, editor_document_dup_title (self));
       break;
@@ -650,6 +739,10 @@ editor_document_set_property (GObject      *object,
 
   switch (prop_id)
     {
+    case PROP_ENCODING:
+      editor_document_set_encoding (self, g_value_get_object (value));
+      break;
+
     case PROP_FILE:
       gtk_source_file_set_location (self->file, g_value_get_object (value));
       break;
@@ -699,6 +792,11 @@ editor_document_class_init (EditorDocumentClass *klass)
                           NULL,
                           (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
+  properties [PROP_ERROR] =
+    g_param_spec_boxed ("error", NULL, NULL,
+                        G_TYPE_ERROR,
+                        (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
   properties [PROP_BUSY] =
     g_param_spec_boolean ("busy",
                           "Busy",
@@ -713,6 +811,13 @@ editor_document_class_init (EditorDocumentClass *klass)
                          -G_MAXDOUBLE, G_MAXDOUBLE, 0.0,
                          (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
+  properties [PROP_ENCODING] =
+    g_param_spec_object ("encoding", NULL, NULL,
+                         EDITOR_TYPE_ENCODING,
+                         (G_PARAM_READWRITE |
+                          G_PARAM_EXPLICIT_NOTIFY |
+                          G_PARAM_STATIC_STRINGS));
+
   properties [PROP_EXTERNALLY_MODIFIED] =
     g_param_spec_boolean ("externally-modified",
                           "Externally Modified",
@@ -741,6 +846,12 @@ editor_document_class_init (EditorDocumentClass *klass)
                          SPELLING_TYPE_CHECKER,
                          (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
+  properties [PROP_STATISTICS] =
+    g_param_spec_object ("statistics", NULL, NULL,
+                         EDITOR_TYPE_DOCUMENT_STATISTICS,
+                         (G_PARAM_READABLE |
+                          G_PARAM_STATIC_STRINGS));
+
   properties [PROP_TITLE] =
     g_param_spec_string ("title",
                          "Title",
@@ -748,6 +859,13 @@ editor_document_class_init (EditorDocumentClass *klass)
                          NULL,
                          (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
+  properties[PROP_NEWLINE_TYPE] =
+    g_param_spec_enum ("newline-type", NULL, NULL,
+                       GTK_SOURCE_TYPE_NEWLINE_TYPE,
+                       GTK_SOURCE_NEWLINE_TYPE_DEFAULT,
+                       (G_PARAM_READABLE |
+                        G_PARAM_STATIC_STRINGS));
+
   g_object_class_install_properties (object_class, N_PROPS, properties);
 
   /**
@@ -764,6 +882,14 @@ editor_document_class_init (EditorDocumentClass *klass)
                                  NULL, NULL,
                                  NULL,
                                  G_TYPE_NONE, 0);
+
+  signals [CURSOR_JUMPED] = g_signal_new ("cursor-jumped",
+                                          G_TYPE_FROM_CLASS (klass),
+                                          G_SIGNAL_RUN_LAST,
+                                          0,
+                                          NULL, NULL,
+                                          NULL,
+                                          G_TYPE_NONE, 0);
 }
 
 static void
@@ -775,6 +901,11 @@ editor_document_init (EditorDocument *self)
   self->file = gtk_source_file_new ();
   self->draft_id = g_uuid_string_random ();
 
+  g_signal_connect (self,
+                    "cursor-moved",
+                    G_CALLBACK (editor_document_cursor_moved),
+                    self);
+
   /* Default implicit-trailing-newline to on in case there are
    * no automatic settings providers.
    */
@@ -1126,8 +1257,7 @@ editor_document_save_cb (GObject      *object,
 
       /* Try again as admin:// */
       if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED) &&
-          g_str_has_prefix (uri, "file://") &&
-          !g_object_get_data (G_OBJECT (saver), "IS_ADMIN"))
+          g_str_has_prefix (uri, "file://"))
         {
           g_autofree char *admin_uri = g_strdup_printf ("admin://%s", uri + strlen ("file://"));
           g_autoptr(GFile) admin_file = g_file_new_for_uri (admin_uri);
@@ -1141,8 +1271,6 @@ editor_document_save_cb (GObject      *object,
           gtk_source_file_saver_set_flags (admin_saver, gtk_source_file_saver_get_flags (saver));
           gtk_source_file_saver_set_newline_type (admin_saver, gtk_source_file_saver_get_newline_type (saver));
 
-          g_object_set_data (G_OBJECT (admin_saver), "IS_ADMIN", GINT_TO_POINTER (TRUE));
-
           gtk_source_file_saver_save_async (admin_saver,
                                             G_PRIORITY_DEFAULT,
                                             NULL,
@@ -1315,7 +1443,10 @@ _editor_document_unmark_busy (EditorDocument *self)
   if (self->busy_count == 0)
     {
       g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_BUSY]);
-      editor_buffer_monitor_unpause (self->monitor);
+
+      if (self->monitor != NULL)
+        editor_buffer_monitor_unpause (self->monitor);
+
       /* Now that we are not busy, emit cursor-moved so that anything that
        * wants to watch the cursor position can safely ignore changes while
        * the document is busy.
@@ -1544,7 +1675,7 @@ editor_document_load_cb (GObject      *object,
 
       g_assert (!file || G_IS_FILE (file));
 
-      self->newline_type = gtk_source_file_loader_get_newline_type (loader);
+      _editor_document_set_newline_type (self, gtk_source_file_loader_get_newline_type (loader));
 
       /* We want to keep the same encoding when saving that was
        * auto-detected from loading the file.
@@ -2186,7 +2317,11 @@ _editor_document_set_encoding (EditorDocument          *self,
 {
   g_return_if_fail (EDITOR_IS_DOCUMENT (self));
 
-  self->encoding = encoding;
+  if (encoding != self->encoding)
+    {
+      self->encoding = encoding;
+      g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_ENCODING]);
+    }
 }
 
 GtkSourceNewlineType
@@ -2203,7 +2338,11 @@ _editor_document_set_newline_type (EditorDocument       *self,
 {
   g_return_if_fail (EDITOR_IS_DOCUMENT (self));
 
-  self->newline_type = newline_type;
+  if (newline_type != self->newline_type)
+    {
+      self->newline_type = newline_type;
+      g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_NEWLINE_TYPE]);
+    }
 }
 
 /**
@@ -2435,3 +2574,54 @@ _editor_document_attach_actions (EditorDocument *self,
                                     "spelling",
                                     G_ACTION_GROUP (self->spell_adapter));
 }
+
+EditorDocumentStatistics *
+editor_document_load_statistics (EditorDocument *self)
+{
+  g_return_val_if_fail (EDITOR_IS_DOCUMENT (self), NULL);
+
+  if (self->statistics == NULL)
+    {
+      self->statistics = editor_document_statistics_new (self);
+      editor_document_statistics_queue_reload (self->statistics);
+    }
+
+  return g_object_ref (self->statistics);
+}
+
+void
+_editor_document_save_insert_mark (EditorDocument *self)
+{
+  GtkTextMark *insert;
+  GtkTextMark *selection_bound;
+  GtkTextIter iter;
+
+  g_return_if_fail (EDITOR_IS_DOCUMENT (self));
+
+  insert = gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (self));
+  selection_bound = gtk_text_buffer_get_selection_bound (GTK_TEXT_BUFFER (self));
+
+  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self), &iter, insert);
+  gtk_text_buffer_move_mark (GTK_TEXT_BUFFER (self), self->saved_insert_mark, &iter);
+
+  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self), &iter, selection_bound);
+  gtk_text_buffer_move_mark (GTK_TEXT_BUFFER (self), self->saved_selection_bound_mark, &iter);
+}
+
+void
+_editor_document_restore_insert_mark (EditorDocument *self)
+{
+  GtkTextIter iter;
+
+  g_return_if_fail (EDITOR_IS_DOCUMENT (self));
+
+  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self), &iter, self->saved_insert_mark);
+  gtk_text_buffer_move_mark (GTK_TEXT_BUFFER (self),
+                             gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (self)),
+                             &iter);
+
+  gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self), &iter, self->saved_selection_bound_mark);
+  gtk_text_buffer_move_mark (GTK_TEXT_BUFFER (self),
+                             gtk_text_buffer_get_selection_bound (GTK_TEXT_BUFFER (self)),
+                             &iter);
+}
